/*
libfawk - A function-only AWK dialect (BASIC compiler)

Copyright (c) 2019..2020 Tibor 'Igor2' Palinkas. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

  * 1. Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
  * 3. Neither the name of the copyright holder nor the names of its
    contributors may be used to endorse or promote products derived from this
    software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Project page: http://repo.hu/projects/libfawk
Source code: svn://repo.hu/libfawk/trunk
Contact the author: http://igor2.repo.hu/contact.html
*/

#include "config.h"
#include "libfawk.h"
#include "parser.h"
#include "fbas.y.tab.h"
#include "lexer_inlines.h"
#include <genht/hash.h>

FAWK_API int fawk_lex_fbas(fbas_STYPE *lval, fawk_ctx_t *ctx)
{
	int chr, nchr;
	char tmp[128];

	restart:;

	if (ctx->parser.in_eof) goto handle_eof;
	lex_textblk_exec(BT_STRING, goto restart, '\r');

	/* token start, leading whitespace already removed */

	ctx->parser.used = 0;

	if (isalpha(chr) || (chr == '_')) { /* read an id or keyword */
		append(chr, return -1);
		for(;;) {
			chr = getch(ctx);
			if (!(isalpha(chr)) && !(isdigit(chr)) && (chr != '_') && (chr != '$')) {
				ungetch(ctx, chr);
				break;
			}
			append(chr, return -1);
			if (chr == '$') { append('\0', return -1); goto got_id; }
		}
		append('\0', return -1); /* append a virtual \0 so it's easier to strcmp */
		if      (genht_strcasecmp(ctx->parser.buff, "if") == 0)        return BT_IF;
		else if (genht_strcasecmp(ctx->parser.buff, "then") == 0)      return BT_THEN;
		else if (genht_strcasecmp(ctx->parser.buff, "else") == 0)      return BT_ELSE;
		else if (genht_strcasecmp(ctx->parser.buff, "for") == 0)       return BT_FOR;
		else if (genht_strcasecmp(ctx->parser.buff, "to") == 0)        return BT_TO;
		else if (genht_strcasecmp(ctx->parser.buff, "step") == 0)      return BT_STEP;
		else if (genht_strcasecmp(ctx->parser.buff, "next") == 0)      return BT_NEXT;
		else if (genht_strcasecmp(ctx->parser.buff, "while") == 0)     return BT_WHILE;
		else if (genht_strcasecmp(ctx->parser.buff, "do") == 0)        return BT_DO;
		else if (genht_strcasecmp(ctx->parser.buff, "loop") == 0)      return BT_LOOP;
		else if (genht_strcasecmp(ctx->parser.buff, "until") == 0)     return BT_UNTIL;
		else if (genht_strcasecmp(ctx->parser.buff, "goto") == 0)      return BT_GOTO;
		else if (genht_strcasecmp(ctx->parser.buff, "def") == 0)       return BT_DEF;
		else if (genht_strcasecmp(ctx->parser.buff, "define") == 0)    return BT_DEF;
		else if (genht_strcasecmp(ctx->parser.buff, "function") == 0)  return BT_DEF;
		else if (genht_strcasecmp(ctx->parser.buff, "end") == 0)       return BT_END;
		else if (genht_strcasecmp(ctx->parser.buff, "and") == 0)       return BT_AND;
		else if (genht_strcasecmp(ctx->parser.buff, "or") == 0)        return BT_OR;
		else if (genht_strcasecmp(ctx->parser.buff, "in") == 0)        return BT_IN;
		else if (genht_strcasecmp(ctx->parser.buff, "let") == 0)       return BT_LET;
		else if (genht_strcasecmp(ctx->parser.buff, "mod") == 0)       return BT_MOD;
		else if (genht_strcasecmp(ctx->parser.buff, "not") == 0)       return BT_NOT;
		else if ((ctx->parser.curr_func != NULL) && (genht_strcasecmp(ctx->parser.buff, ctx->parser.curr_func) == 0)) return BT_CURRFUNC;
		else if (genht_strcasecmp(ctx->parser.buff, "rem") == 0)       { readtil(ctx, "\n"); goto restart; }
		else if (genht_strcasecmp(ctx->parser.buff, "include") == 0)   { lex_include(goto restart); }
		else {
			got_id:; lval->str = fawk_strdup(ctx, ctx->parser.buff);
			return (lval->str == NULL) ? -1 : BT_ID;
		}
	}
	else if (isdigit(chr)) { /* read a number */
		append(chr, return -1);
		if (chr == '0') {
			nchr = getch(ctx);
			if (nchr == 'x') {
				fawk_readup(ctx, "0123456789abcdefABCDEF");
				lval->num = strtol(ctx->parser.buff, NULL, 16);
				return BT_NUM;
			}
			ungetch(ctx, nchr);
		}
		return read_numeric(ctx, &lval->num, 0, BT_NUM);
	}
	else switch(chr) {
		case '@': case '?': case ':': case '(': case ')':
		case '\\': case ']': case ',': case '{': case '}':
		case '+': case '-': case '*': case '/': case '%': case '=': case '\n':
			return chr;

		case '[': { lex_textblk_start(goto restart); }

		case '.':
			nchr = getch(ctx);
			if (!isdigit(nchr)) {
				ungetch(ctx, nchr);
				return '.';
			}
			else {
				append('.', return -1); append(nchr, return -1);
				return read_numeric(ctx, &lval->num, 1, BT_NUM);
			}
			break;

		case '<':
			nchr = getch(ctx); append(chr, return -1);
			switch(nchr) {
				case '=': return BT_LTEQ;
				case '>': return BT_NEQ;
				default: ungetch(ctx, nchr); return chr;
			}
			break;

		case_op_ch2('>', '=', BT_GTEQ);

		case '\"':
			if (read_strlit(ctx, '\"') == 0) return BT_NIL;
			lval->str = fawk_strdup(ctx, ctx->parser.buff);
			return (lval->str == NULL) ? -1 : BT_STRING;

		case EOF:   { lex_got_eof(goto restart); }

		default:
			sprintf(tmp, "Invalid character on input: '%c'\n", chr);
			LIBFAWK_ERROR(ctx, tmp, ctx->parser.isp->fn, ctx->parser.isp->line+1, ctx->parser.isp->col+1,  -1);
	}
}
